import sys
import traceback
import yaml
from cStringIO import StringIO
from timeit import default_timer
from collections import defaultdict

from django.core.management import call_command, CommandError
from django.core.management.base import BaseCommand
from django.db.utils import DatabaseError


class MigrationSession(object):
    """
    Class which is initialized with Django app/model migrations to perform.
    Performs migrations while keeping track of the state of each migration.
    Provides the state of all migrations on demand.
    """
    def __init__(self, input_yaml, stderr):
        self.to_apply = []
        self.migration_state = {
            'success': [],
            'failure': None,
            'unapplied': [],
            'rollback_commands': [],
        }
        self.timer = default_timer
        self.stderr = stderr

        # Load the passed-in YAML into a dictionary.
        self.input_migrations = yaml.safe_load(input_yaml)

        # Build a list of migrations to apply in order.
        for migration in self.input_migrations['migrations']:
            self.to_apply.append(migration)

    def more_to_apply(self):
        """
        Returns True if more migrations remain to apply in this session, else False.
        """
        return len(self.to_apply) > 0

    def _add_rollback_commands(self):
        """
        Generate rollback commands for the apps that have had migrations applied.
        If an app's migration has failed, include that rollback command as well.
        """
        apps_to_rollback = set()
        # Add the apps with successfully applied migrations.
        apps_to_rollback.update([m['migration'][0] for m in self.migration_state['success']])
        # If an app migration failed, include that rollback also.
        if self.migration_state['failure']:
            apps_to_rollback.add(self.migration_state['failure']['migration'][0])
        for app in apps_to_rollback:
            initial_app_state = None
            for initial in self.input_migrations['initial_states']:
                if app == initial[0]:
                    initial_app_state = initial
                    break
            if not initial_app_state:
                raise CommandError('App "{}" not found in initial migration states.'.format(app))
            self.migration_state['rollback_commands'].append(
                [
                    'python', 'manage.py', 'migrate',
                    app,
                    initial_app_state[1]
                ]
            )

    def _apply_next(self):
        """
        Applies the next-in-line Django model migration.
        """
        if not self.more_to_apply():
            return

        app, migration = self.to_apply.pop(0)

        out = StringIO()
        start = self.timer()
        try:
            call_command("migrate", app_label=app, migration_name=migration, noinput=True, stdout=out)
        except (CommandError, DatabaseError) as e:
            time_to_fail = self.timer() - start
            exc_type, exc_value, exc_traceback = sys.exc_info()
            # Assumed that only a single migration failure will occur.
            self.migration_state['failure'] = {
                'migration': [app, migration],
                'duration': time_to_fail,
                'output': out.getvalue(),
                'traceback': repr(traceback.format_exception(exc_type, exc_value, exc_traceback)),
            }
            # Add the remaining migrations to the unapplied status.
            while self.more_to_apply():
                app_migration = self.to_apply.pop(0)
                self.migration_state['unapplied'].append(app_migration)
            # Find the apps that have been applied -or- failed.
            # Include their initial migrations as commands.
            self._add_rollback_commands()
            raise CommandError("Migration failed for app '{}' - migration '{}'.\n".format(app, migration))

        time_to_apply = self.timer() - start
        self.migration_state['success'].append({
            'migration': [app, migration],
            'duration': time_to_apply,
            'output': out.getvalue(),
        })

    def apply_all(self):
        """
        Apply all the migrations.
        """
        while self.more_to_apply():
            self._apply_next()
        self._add_rollback_commands()

    def state(self):
        """
        Returns the current state as a YAML string.
        """
        return yaml.dump(self.migration_state)


class Command(BaseCommand):
    """
    Given a YAML file containing apps and migrations, apply those app migrations.
    The YAML format for the apps/migrations is:

    migrations:
      - [app1, 0001_initial]
      - [app2, 0012_otherthing]
      - [app1, 0002_somthing]
    initial_states:
      - app1:
        - zero
      - app2:
        - 0011_beforetheotherthing

    If an error occurs in any of the migrations, the migrations are halted at that point
        and the status is recorded in an artifact.
    The output of the command is in YAML format and specifies the following:
        - migrations that run successfully and how long they took
        - migrations that failed and how long the failure took
        - migrations that unapplied due to previous failures
    The output YAML format:

    success:
    - migration: [app1, 0001_initial]
      duration: 3.45
      output: All good!
    failure:
      migration: [app2, 0012_otherthing]
      output: This migration has failed!!!!
    unapplied:
    - [app1, 0002_something]
    rollback_commands:
    - [python, manage.py, migrate, app1, zero]

    If an output file is specified, the YAML output is also directed to that file.

    Rollbacks due to migration failures are left to the mgmt command user.
    """
    help = "Given a YAML file containing apps and migrations, apply those app migrations."

    def add_arguments(self, parser):
        parser.add_argument(
            'input_file',
            type=unicode,
            nargs='?',
            help="Filename from which apps/migrations will be read."
        )
        parser.add_argument(
            '--output_file',
            dest='output_file',
            default=None,
            help="Filename to which migration results will be written."
        )

    def handle(self, *args, **kwargs):
        with open(kwargs['input_file'], 'r') as f:
            input_yaml = f.read()
        migrator = MigrationSession(input_yaml, self.stderr)

        failure = False
        try:
            migrator.apply_all()
        except CommandError as e:
            self.stderr.write("Migration error: {}".format(e.message))
            failure = True

        self.stdout.write(migrator.state())
        if kwargs['output_file']:
            with open(kwargs['output_file'], 'w') as outfile:
                outfile.write(migrator.state())

        sys.exit(int(failure))



import re
import sys
import json
from cStringIO import StringIO
from collections import defaultdict
import yaml

from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db.utils import OperationalError
from django.db.migrations.loader import MigrationLoader
from django.db import DEFAULT_DB_ALIAS, connections

class Command(BaseCommand):
    """
    Checks for unapplied migrations.
    Prints out a YAML string of any unapplied migrations, along with their accompanying application name *and*
    the initial migration state of every application with unapplied migrations.
    For example:

    migrations:
      - [app1, 0001_initial]
      - [app2, 0012_otherthing]
      - [app1, 0002_somthing]
    initial_states: {app1: zero, app2: 0011_thisthing}

    If all migrations are applied, returns an empty YAML "migrations" dict.
    This command can be used in a couple of ways:
    1) To generate a list of unapplied migrations
    2) To detect if any unapplied migrations exist and failing if so (by specifying '--fail_on_unapplied')
    """
    help = "Prints out a YAML string of any unapplied migrations, along with their accompanying application name."

    def add_arguments(self, parser):
        parser.add_argument(
            '--fail_on_unapplied',
            dest='fail_on_unapplied',
            action='store_true',
            help="If flag specified, command will exit with a non-zero value when unapplied migrations exist.",
        )
        parser.add_argument(
            '--database',
            dest='database',
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to synchronize. Defaults to the "default" database.',
        )
        parser.add_argument(
            '--output_file',
            dest='output_file',
            default=None,
            help="Filename to which output should be written."
        )

    def _gather_unapplied_migrations(self, loader):
        """
        Output a dictionary of unapplied migrations in the form {'app1': ['migration1', migration2']}.
        This implementation is mostly copied from the Django 'showmigrations' mgmt command.
        https://github.com/django/django/blob/stable/1.8.x/django/core/management/commands/showmigrations.py
        """
        unapplied = []
        graph = loader.graph
        plan = []
        seen = set()

        # Generate the plan, in the order that migrations have been/should be applied.
        for target in graph.leaf_nodes():
            for migration in graph.forwards_plan(target):
                if migration not in seen:
                    plan.append(graph.nodes[migration])
                    seen.add(migration)

        # Remove the migrations that have already been applied.
        for migration in plan:
            if not (migration.app_label, migration.name) in loader.applied_migrations:
                # NOTE: Unicode Django application names are unsupported.
                unapplied.append([migration.app_label, str(migration.name)])
        return unapplied

    def _gather_current_migration_state(self, loader, apps):
        """
        Extract the most recent migrations from the relevant apps.
        If no migrations have been performed, return 'zero' as the most recent migration for the app.
        """
        # Only care about applied migrations for the passed-in apps.
        apps = set(apps)
        relevant_applied = [migration for migration in loader.applied_migrations if migration[0] in apps]
        # Sort them by the most recent migration and convert to a dictionary,
        # leaving apps as keys and most recent migration as values.
        most_recents = dict(sorted(relevant_applied, key=lambda m: m[1], reverse=True))
        # Fill in the apps with no migrations with 'zero'.
        # NOTE: Unicode Django application names are unsupported.
        most_recents = [[app, 'zero' if not app in most_recents else str(most_recents[app])] for app in apps]
        return most_recents

    def _gather_migration_info(self, *args, **kwargs):
        """
        Return:
            - the unapplied migrations from each app
            - the initial migration state of any app with unapplied migrations
        """
        db = kwargs['database']
        connection = connections[db]
        loader = MigrationLoader(connection, ignore_no_migrations=True)
        unapplied = self._gather_unapplied_migrations(loader)
        currents = self._gather_current_migration_state(loader, [u[0] for u in unapplied])
        return {'migrations': unapplied, 'initial_states': currents}

    def handle(self, *args, **kwargs):
        migration_info = self._gather_migration_info(self, *args, **kwargs)

        # Compose the output YAML.
        yaml_output = yaml.dump(migration_info)

        # Output the composed YAML.
        self.stdout.write(yaml_output)
        if kwargs['output_file']:
            with open(kwargs['output_file'], 'w') as outfile:
                outfile.write(yaml_output)

        if kwargs['fail_on_unapplied'] and migration_info['migrations']:
            sys.exit(1)
        else:
            sys.exit(0)



#!/usr/bin/env python
import io
from setuptools import setup, find_packages


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement;
    that is, it is not blank, a comment, or editable.
    """
    # Remove whitespace at the start/end of the line
    line = line.strip()

    # Skip blank lines, comments, and editable installs
    return not (
        line == '' or
        line.startswith('-r') or
        line.startswith('#') or
        line.startswith('-e') or
        line.startswith('git+')
    )


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.
    Returns a list of requirement strings.
    """
    requirements = set()
    for path in requirements_paths:
        requirements.update(
            line.strip() for line in open(path).readlines()
            if is_requirement(line)
        )
    return list(requirements)


long_description = io.open('README.md', encoding='utf-8').read()

METADATA = dict(
    name='edx-django-release-util',
    version='0.0.1',
    description='edx-django-release-util',
    author='edX',
    author_email='oscm@edx.org',
    long_description=long_description,
    license='AGPL',
    url='http://github.com/edx/edx-django-release-util',
    install_requires=load_requirements('requirements/base.txt',),
    tests_require=load_requirements('requirements/test.txt'),
    packages=find_packages(exclude=['*.test', '*.tests']),
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Environment :: Web Environment',
        'Topic :: Internet',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django',
    ],
)

if __name__ == '__main__':
    setup(**METADATA)
